/*
   Module .............libbt                                
   File ...............SimpleControl.h                                          
   Author .............Traveler Hauptman                                 
   Creation Date ......Nov 24, 2002                                 
                                                                       
              Copyright (c) 2002 Barrett Technology, Inc.              
                         139 Main Street                               
                        Kendall/MIT Square                             
                   Cambridge, MA  02142-1528  USA                      
                                                                      
                         All rights reserved.                          
                                                                       
 ***********************************************************************
 *                            DISCLAIMER                               *
 *                                                                     *
 *  This software and related documentation are provided to you on     *
 *  an as is basis and without warranty of any kind.  No warranties,   *
 *  express or implied, including, without limitation, any warranties  *
 *  of merchantability or fitness for a particular purpose are being   *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett    *
 *  Technology, Inc. be liable for any lost development expenses, lost *
 *  lost profits, or any incidental, special, or consequential damage. *
 *                                                                     *
 ***********************************************************************  
 *                                                                      
 *  NOTES:                                                              
 *   A simple Torque, PID & Trapezoidal velocity trajectory controller                        
 *                                                                      
 *  REVISION HISTORY:                                                   
 *  021124 - TH - File created                                          
 *                                                                      
 *======================================================================*/

 /* \file SimpleControl.h  
    \brief A simple Torque, PID & Trapezoidal velocity trajectory controller
    
    
*/ 
#ifndef _SIMPLECONTROL_H
#define _SIMPLECONTROL_H
#ifdef __cplusplus
extern "C"
{
#endif/* __cplusplus */
/*================================================PID stuff================================*/

#include <sys/neutrino.h>

#include <pthread.h>

#ifndef PI
#define PI 3.141596
#endif PI
#define		MAX_FILT_SIZE  5		   /* max size of filter coefficient arrays */
#define		FIRST_ORDER 	1
#define		SECOND_ORDER   2
#define		FOURTH_ORDER   4

/* Define the Filter structure */
typedef struct
{
  int   order;             // FIRST_ORDER, SECOND_ORDER, OR FOURTH_ORDER
  double  cutoffHz;          // cutoff frequency in Hz
  double  zeta;        // damping factor (between 0 and 1)
  double  d[MAX_FILT_SIZE];  // coefficients of denominator
  double  n[MAX_FILT_SIZE];  // coefficients of numerator
  double  x[MAX_FILT_SIZE];  // old input values
  double  y[MAX_FILT_SIZE];  // old output values
    int   index;         // index of current value in input and output arrays
} filter_struct;
double  Filter (filter_struct *filt, double xnew);
void  InitDifferentiator (filter_struct *filt, double sample_time);
void InitPassthroughFilter (filter_struct *filt);

/** PID regulator

PIDregulator maintains configuration and state information for a generic PID controller.

Usage:
  -# declare a PIDregulator variable (ex: PIDregulator pid;)
  -# Use PIDreset() to set accumulated error to zero and yref to y
  -# In your control loop:
  - Set pid.y to the present value of y;
  - Use PIDcalc() to calculate the result
  
*/
typedef struct  
{
  double Kp; //!< Proportional gain
  double Kd; //!< Derivative gain
  double Ki; //!< Integral gain

  double e;  //!< Error (yref - y)
  double de; //!< Derivative of error
  double fe;
  double se; //!< Sum of error
  
  double laste;
  int firsttick;   //!< handles special case de for first tick

  double dt;        //!< Defaults to 1
  double saturation; //!< Defaults to zero (doesn't ever saturate)

  double y; //!< Measured control parameter
  double yref; //!< Commanded control parameter
  double lastresult; //!< Last output of the regulator
  int filter_de; //!< Should we use a filter on de?
  filter_struct de_filter,se_filter;
}PIDregulator;

void PIDinit(PIDregulator *pid,double Kp, double Kd, double Ki, double dt);
void PIDmeasured(PIDregulator *pid, double y);
void PIDcommand(PIDregulator *pid, double yref);
void PIDreset(PIDregulator *pid); 
double PIDcalc(PIDregulator *pid);

/*================================================Trajectory stuff================================*/

enum {TRAJ_DONE,TRAJ_RUN,TRAJ_FILE};

typedef struct
{
  double x,v,vel,acc;
}trj_leg_struct;

typedef struct
{
  double x,v,t;
}trj_corner_struct;

typedef struct
{
  trj_leg_struct *point;
  int num;
}point_list_struct;


/*  Trajectory generator (this is an aborted attempt at a smooth trajectory generator. Left in in case it's needed later)

  trajectory provides configuration and state information for a set of trapezoidal tragectory
  generation functions specified in trajectory.c.
  
  Make sure you set acc, and vel. These are the constant acceleration used and the 
  maximum velocity allowed respectively.

*/
typedef struct 
{
  int state; //!< 0: done, 1:run, 2: running crippled
  point_list_struct pl;
  int current_point;
  
  int loop; // 0 don't loop, 1 do loop
  int phase; // start, coast, end
  int crippled;
  
  double cmd;
  double vel;
  double extra_t;
  trj_corner_struct c0,c1,c2,c3;

}npoint_traj;



/*! \brief Trajectory generator

  trajectory provides configuration and state information for a set of trapezoidal tragectory
  generation functions specified in trajectory.c.
  
  Make sure you set acc, and vel. These are the constant acceleration used and the 
  maximum velocity allowed respectively.

*/

typedef struct 
{
  int state; //!< 0: done, 1:run
  double cmd;
  double start_point;
  double sign;
  double end_point;
  double end;
  double acc;
  double vel;
  double t;
  double t1,t2; // Calculated inflection points of trapezoidal velocity curve
  double x1,x2;
  double **tf; //Pointer to a trajectory file
  int row; //Present row of the trajectory file
  int rows; //Total number of rows in trajectory file 
}trajectory;

double evaluate_trajectory(trajectory *traj,double dt);
void init_trajectory(trajectory *traj, double start, double end);
double calc_traj_time(trajectory *traj, double start, double end, double vel, double acc);
void sync_trajectory(trajectory *traj, double start, double end, double vel, double acc, double t);
double scale_vel(trajectory *traj, double start, double end, double vel, double acc, double t);
double scale_acc(trajectory *traj, double start, double end, double vel, double acc, double t);

typedef struct {
  double vel,acc1,acc2,dt_vel,dt_acc1,dt_acc2;
}Seg_int;

typedef struct 
{
  int state;
  double t;
  int col; //column of point list table
  int idx,n;
  double Qdot,Qdot_next;
  double v_prev,v_next;
  Seg_int seg;
  
  
  //===========parameters
  double trj_acc;
  
  
  //===========state
  int segment; //0 = acc, 1=vel
  double dt_acc,dt_vel; 
  double t_acc, t_vel, q_acc, q_vel;
  double t0,q0,v0;
  double acc,vel;
  double last_vel;
  
  
  double **tf; //Pointer to a trajectory file
  int row; //Present row of the trajectory file
  int rows; //Total number of rows in trajectory file 
     
  
}via_trj;



double linterp(double x1, double y1, double x2, double y2, double x);
double eval_via_trj(via_trj *trj,double dt);
void start_via_trj(via_trj *trj,double start,int n,int col,double acc);
void CalcSegment(Seg_int *seg,double q1, double q2, double t1, double t2, double v_prev, double v_next, double seg_acc, int end);

/*================================================Controller stuff================================*/
/*! \brief Simple torque, pid, and trapezoidal trajectory controller

This structure stores state information along with pid and trajectory info for
a simple motor controller. The functions that use this structure are defined in
SimpleControl.c.

*/
typedef struct
{
  int mode; //!< 0:idle, 1:torque, 2:pid 3:vel 4:trj
  double requested_torque;
  double command_torque;
  double kill_torque;
  double last_dt;
  double position;
  PIDregulator pid;
  trajectory trj;
  via_trj via;
  int error; //nonzero if there are any errors
  double ddebug;
  int idebug;
  pthread_mutex_t mutex;
}SimpleCtl;

int SCinit(SimpleCtl *sc);
double SCevaluate(SimpleCtl *sc, double position, double dt);
int SCsetmode(SimpleCtl *sc, int mode);
int SCsettorque(SimpleCtl *sc,double torque);
int SCsetpid(SimpleCtl *sc,double kp,double kd,double ki, double saturation);//!!implement saturation variable
int SCsetcmd(SimpleCtl *sc,double cmd);
int SCsettrjprof(SimpleCtl *sc,double vel,double acc);
int SCstarttrj(SimpleCtl *sc,double end);
int SCfindLimit(SimpleCtl *sc,double offset); //!!trajectory to opposite stop until kill_trq, then to this stop
int SCstartTrjFile(SimpleCtl *sc, double ***trajFile, int rows);
int SCstartViaTrjFile(SimpleCtl *sc, double ***trajFile, int n,int col,double timescale);

#ifdef __cplusplus
}
#endif/* __cplusplus */
#endif _SIMPLECONTROL_H


/*======================================================================*
 *                                                                      *
 *             Copyright (c) 2003 Barrett Technology, Inc.              *
 *                        139 Main Street                               *
 *                       Kendall/MIT Square                             *
 *                  Cambridge, MA  02142-1528  USA                      *
 *                                                                      *
 *                        All rights reserved.                          *
 *                                                                      *
 *  ******************************************************************  *
 *                            DISCLAIMER                                *
 *                                                                      *
 *  This software and related documentation are provided to you on      *
 *  an as is basis and without warranty of any kind.  No warranties,    *
 *  express or implied, including, without limitation, any warranties   *
 *  of merchantability or fitness for a particular purpose are being    *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett     *
 *  Technology, Inc. be liable for any lost development expenses, lost  *
 *  lost profits, or any incidental, special, or consequential damage.  *
 *  ******************************************************************  *
 *                                                                      
 * CVS info: $Id$
 * CVS automatic log (prune as desired):
 * $Log$
 *
 *======================================================================*/



