/*
   Module .............libbt                                
   File ...............control_loop.c                                     
   Author .............Traveler Hauptman                                 
   Creation Date ......Nov 23, 2002                                 
                                                                       
              Copyright (c) 2002 Barrett Technology, Inc.              
                         139 Main Street                               
                        Kendall/MIT Square                             
                   Cambridge, MA  02142-1528  USA                      
                                                                      
                         All rights reserved.                          
                                                                       
 ***********************************************************************
 *                            DISCLAIMER                               *
 *                                                                     *
 *  This software and related documentation are provided to you on     *
 *  an as is basis and without warranty of any kind.  No warranties,   *
 *  express or implied, including, without limitation, any warranties  *
 *  of merchantability or fitness for a particular purpose are being   *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett    *
 *  Technology, Inc. be liable for any lost development expenses, lost *
 *  lost profits, or any incidental, special, or consequential damage. *
 *                                                                     *
 ***********************************************************************  
 *                                                                      
 *  NOTES:                                                              
 *   Infrastructure to simplify adding and using your own control threads. 
 *                                                                      
 *  REVISION HISTORY:                                                   
 *  021123 - TH - File created                                          
 *                                                                      
 *======================================================================*/

 /*! \file control_loop.c
    \brief Infrastructure to simplify adding and using your own control threads.
    
    This module adds code to start up timing threads that will call a user specified 
  control thread at regular intervals.
    
   
*/

/*==============================*
 * INCLUDES - System Files      *
 *==============================*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <syslog.h>
#include <stddef.h>
#include <signal.h>
#include <pthread.h>
#include <errno.h>
#include <semaphore.h>
#include <malloc.h>
#include <process.h>
#include <time.h>
#include <inttypes.h>
#include <sys/sched.h>
#include <sys/mman.h>
#include <sys/neutrino.h>
#include <sys/netmgr.h>
#include <sys/syspage.h>

/*==============================*
 * INCLUDES - Project Files     *
 *==============================*/
#include "control_loop.h"
#include "globals.h"
#include "btcan.h"
#include "btsystem.h"
#include "traj.h"

sem_t     timer_semaphore;
pthread_t timer_thd_id;
pthread_t control_thd_id;
int shutdown_threads=1;
uint64_t ms_timer,sem_cntr;
long sample_period2;

/** Creates a timing thread and links it to your control loop function 

  \param priority The priority to run the timer and control loop threads at.
  \param sample_period An integer to multiply the system clock base epoch by. 
  If an integer multiple is not used, there will be problems of missing control 
  loops.

*/
void start_control_threads(int priority, long sample_period, void *function,void *args)
{

  pthread_attr_t       timer_attr;
  pthread_attr_t       control_attr;
  struct sched_param   timer_param;
  struct sched_param   control_param;
  struct _clockperiod  clock_period;
  
  if (!shutdown_threads) return;
  
  
   //Set the resolution of the system clock
  clock_period.nsec = EPOCH;
  clock_period.fract = 0;
  if(ClockPeriod(CLOCK_REALTIME,&clock_period,NULL,0) == -1)
  {
    syslog(LOG_ERR,"Failed to initialize clockperiod: %m");
    exit(-1);
  }
  
  sample_period2 = sample_period;
  
  shutdown_threads = 0;
  if ((sem_init(&timer_semaphore, 1, 0)) == -1)
  {
    syslog(LOG_ERR,"start_control_threads:Semaphore initialization failure!");
    exit(-1);
  }
    /* start timer_thread - the thread that posts the timer semaphore to the control_thread */

  pthread_attr_init(&timer_attr);
  pthread_attr_setinheritsched(&timer_attr, PTHREAD_EXPLICIT_SCHED);
  timer_param.sched_priority = priority;
  pthread_attr_setschedparam(&timer_attr, &timer_param);

  pthread_create(&timer_thd_id, &timer_attr, (void*)TimerThread, 0);
  sched_setscheduler(timer_thd_id, SCHED_FIFO, &timer_param);
  if (timer_thd_id == -1)
  {
    syslog(LOG_ERR,"start_control_threads:Couldn't start timer thread!");
    exit(-1);
  }

    /* start control_thread - the thread which controls the robot */
  pthread_attr_init(&control_attr);
  pthread_attr_setinheritsched(&control_attr, PTHREAD_EXPLICIT_SCHED);
  control_param.sched_priority = priority;
  pthread_attr_setschedparam(&control_attr, &control_param);

  pthread_create(&control_thd_id, &control_attr, function, args);
  sched_setscheduler(control_thd_id, SCHED_FIFO, &control_param);

  if (control_thd_id == -1)
  {
    syslog(LOG_ERR,"start_control_threads:Couldn't start control thread!");
    exit(-1);
  }
  
  StartRecordingThread(); //Start the recording thread
  
}
/** Stops the control threads

*/
void stop_control_threads()
{
  
 shutdown_threads = 1;
 usleep(10000);
 sem_destroy(&timer_semaphore); 
}

void TimerThread()
{
  struct sched_param param;
  int                timer_channel_id;
  timer_t            timer_id;
  struct sigevent    timer_event;
  struct itimerspec  timer_specs;
  struct _pulse      timer_pulse;
  int                timer_connection_id;
  int                value;
  int oldtype;
  
//  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype ); //Cancelation can happen anywhere
//  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,&oldtype );       //Enable cancellation


  /*Create a channel for the timer client to pulse us on*/
  if ((timer_channel_id = ChannelCreate(0)) == -1)
  {
    syslog(LOG_ERR,"startSemaphoreThread:Unable to create channel!");
    exit(-1);
  }
  if ((timer_connection_id = ConnectAttach(0, 0, timer_channel_id, _NTO_SIDE_CHANNEL, 0)) < 0)
  {
    syslog(LOG_ERR,"startSemaphoreThread:Unable to attach channel!");
    exit(-1);
  }

  /*load event structure with pulse info */
  SIGEV_PULSE_INIT(&timer_event, timer_connection_id, getprio(0), _PULSE_CODE_MINAVAIL, NULL);

  /* Create the timer, and tie it to the pulse event from above */
  timer_create(CLOCK_REALTIME, &timer_event, &timer_id);

  timer_specs.it_value.tv_sec   = 0L;
  timer_specs.it_value.tv_nsec  = SAMPLETIME*sample_period2;
  timer_specs.it_interval.tv_sec  = 0L;
  timer_specs.it_interval.tv_nsec = SAMPLETIME*sample_period2;

  /* Settime starts the timer.  And with the it_interval non-zero, it will refire continuously */
  if (timer_settime(timer_id, 0, &timer_specs, NULL) == -1)
  {
    syslog(LOG_ERR,"startSemaphoreThread:Unable to set semaphore timer interval!\n");
    syslog(LOG_ERR,"Error #: %d, %m\n", errno);
    exit(-1);
  }
  ms_timer = 0;
  sem_cntr = 0;
  
  while(!shutdown_threads)
  {
      
    if(MsgReceive(timer_channel_id, &timer_pulse, sizeof(struct _pulse), NULL) < 0)
    {
      syslog(LOG_ERR,"startSemaphoreThread:Pulse size not big enough for a pulse.");
      exit(-1);
    }
    
    if(timer_pulse.code == _PULSE_CODE_MINAVAIL)
    {
      ms_timer++;
      sem_cntr++;
      if (sem_getvalue(&timer_semaphore,&value) == 0)
      {
        if(value < 1)
        {
          sem_cntr--;
          if ((sem_post(&timer_semaphore)) == -1)
          {
            syslog(LOG_ERR,"startSemaphoreThread:Unable to post semaphore!");
          }
        }

      }
    }
  }
  
  timer_delete(timer_id);
  ConnectDetach(timer_connection_id);
  ChannelDestroy(timer_channel_id);
  pthread_exit(NULL);
}
/** Example control loop for use with start_control_threads() Do not use this! */
void ControlThread(void *data)
{
  int num_actuators;
  actuator_struct *act;
  SimpleCtl *sc;
  int torques[5] = {0,0,0,0,0};
  long int position;
  
  num_actuators = ((control_thd_parms *)data)->num_act;
  act = ((control_thd_parms *)data)->act;
  sc = ((control_thd_parms *)data)->sc;

  
  
  while(!shutdown_threads)
  {
    sem_wait(&timer_semaphore);


  
    //TimedPuckGet(act->P.puck_ID,ACTUAL_POSITION,&position,10000);  
    
    
  }
  
  pthread_exit(NULL);
}

/*======================================================================*
 *                                                                      *
 *             Copyright (c) 2003 Barrett Technology, Inc.              *
 *                        139 Main Street                               *
 *                       Kendall/MIT Square                             *
 *                  Cambridge, MA  02142-1528  USA                      *
 *                                                                      *
 *                        All rights reserved.                          *
 *                                                                      *
 *  ******************************************************************  *
 *                            DISCLAIMER                                *
 *                                                                      *
 *  This software and related documentation are provided to you on      *
 *  an as is basis and without warranty of any kind.  No warranties,    *
 *  express or implied, including, without limitation, any warranties   *
 *  of merchantability or fitness for a particular purpose are being    *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett     *
 *  Technology, Inc. be liable for any lost development expenses, lost  *
 *  lost profits, or any incidental, special, or consequential damage.  *
 *  ******************************************************************  *
 *                                                                      
 * CVS info: $Id$
 * CVS automatic log (prune as desired):
 * $Log$
 *
 *======================================================================*/
  
 
